# js dec2024 cleaned up code with vulture,
#    updated library calls for OSMnx version 2.0.0
#    updated library calls for Folium 0.19.2

import osmnx as ox
import networkx as nx
import folium
import pickle
import time, datetime
import os.path
import os as os
import webbrowser

ox.settings.log_console=True
ox.settings.use_cache=True

print('OSMnx version: ', ox.__version__)
#tested against OSMnx version:  1.7.0
print('NetworkX version: ',nx.__version__)

start_time = time.time()

def elapsed_time(start):
    end_time = time.time()
    duration = round(end_time - start)
    timestamp = str(datetime.timedelta(seconds = duration))
    print(timestamp)

# Using network_type='walk' seems to include dirt roads, so use a custom filter
cf = '["highway"~"path|footway"]'
cf_roads = '["highway"~"motorway|motorway_link|trunk|primary|secondary|tertiary|unclassified|residential|living_street|residential|service|track"]'
cf_add_routes = '["highway"~"path|footway|motorway|motorway_link|trunk|primary|secondary|tertiary|unclassified|residential|living_street|residential|service|track"]'

n,s,e,w = 33.92485, 32.73704, -107.65983, -108.94574 # rough box of Gila NF

gila_trail_filename = 'gila from bbox.pkl'
gila_road_filename =  'gila roads from bbox.pkl'

G  = nx.MultiDiGraph()
GR = nx.MultiDiGraph()

# Because our program development needs a great deal of iteration,
# store intermediate results in pickle files
if not os.path.exists(gila_trail_filename):
    G =  ox.graph_from_bbox((w,s,e,n),simplify=False,retain_all=True,custom_filter=cf)
    GR = ox.graph_from_bbox((w,s,e,n),simplify=False,retain_all=True,custom_filter=cf_roads)

    pickle.dump(G,open(gila_trail_filename, 'wb'))
    pickle.dump(GR,open(gila_road_filename, 'wb'))
    print('finished importing trail and road graphs')
    elapsed_time(start_time)
else:    
    G =  pickle.load(open(gila_trail_filename, 'rb'))
    GR = pickle.load(open(gila_road_filename, 'rb'))
    print('finished loading trail and road graphs')
    elapsed_time(start_time)

  

def draw(Q,title='',show_nodes = False, boxes=None, not_boxes = None, markers=None):
    start_latlng = (33.04876, -108.30680)
    m = folium.Map(location=[start_latlng[1],start_latlng[0]],
               zoom_start=12)
    # convert to MultiDiGraph for folium
    QD = Q.to_directed()
    QD.graph['crs'] = ox.settings.default_crs

    # our graphs may be HUGE, so simplify if not already done so
    if not "simplified" in QD.graph:
        QD = ox.simplify_graph(QD, remove_rings=True, track_merged=True)
    
    nodes, edges = ox.graph_to_gdfs(QD)
    m = edges.explore(color='#ff00ff', tiles="openstreetmap")
    if show_nodes:    
        nodes.explore(m=m, color="pink", marker_kwds={"radius": 4})    

    if not boxes == None:
        for n,s,e,w in boxes:
            folium.Rectangle(bounds=[(n,e),(s,w)], color='#ff7800', fill=True, fill_color='#ffff00', fill_opacity=0.2).add_to(m)
    if not not_boxes == None:
        for n,s,e,w in not_boxes:
            folium.Rectangle(bounds=[(n,e),(s,w)], color='#00ff78', fill=False, fill_color='#00ffff', fill_opacity=0.2).add_to(m)
    if not markers == None:
        for lat,lon, txt in markers:
            #print(lat,lon,txt)
            folium.Marker(location=[lat,lon],popup=txt,
                          tooltip=txt).add_to(m)
            
    # include the option to switch map layers
    folium.TileLayer('openstreetmap').add_to(m)

    folium.TileLayer('cartodbpositron').add_to(m)
#    folium.TileLayer('Stamen Terrain').add_to(m)
    folium.TileLayer(
        tiles='https://{s}.tile-cyclosm.openstreetmap.fr/cyclosm/{z}/{x}/{y}.png',
        name='cyclosm',
        attr='CyclOSM',
    ).add_to(m)
    folium.LayerControl().add_to(m)    

    title_html = '''
            <h3 align="center" style="font-size:16px"><b>{}</b></h3>
             '''.format(title)   
    m.get_root().html.add_child(folium.Element(title_html))
    file_name = "/home/sagebrush/biz/graph/map_" + title + ".html"
    m.save(file_name)
#    m.show_in_browser()
    webbrowser.open(file_name)
    input("Press Enter to continue...")
    print('*************', flush=True)
    print("", flush=True)    
    


# ---------- mind the gap section --------    

def distance(Q, node, node2): #assumes we are in polar coordinates
    y1 = Q.nodes[node]['y']
    x1 = Q.nodes[node]['x']
    y2 = Q.nodes[node2]['y']
    x2 = Q.nodes[node2]['x']
    #d = ox.distance.euclidean(y1, x1, y2, x2)
    # we are in lat/long, so use different function
    d = ox.distance.great_circle(y1,x1, y2, x2) # should return distance in meters
    return d

def distance_road(Q,node,QR):
    y = Q.nodes[node]['y']
    x = Q.nodes[node]['x']
    noder, d = ox.distance.nearest_nodes(QR, x, y,  return_dist=True)
    return d        

def connect_trailheads_with_road(Q,node,node2, margin=100):
    y1 = Q.nodes[node]['y']
    x1 = Q.nodes[node]['x']
    y2 = Q.nodes[node2]['y']
    x2 = Q.nodes[node2]['x']
    ymargin = abs(y1-y2)/8.0
    xmargin = abs(x1-x2)/8.0
    margin = xmargin + ymargin
    
    n = max(y1,y2) + margin
    s = min(y1,y2) - margin
    e = max(x1,x2) + margin
    w = min(x1,x2) - margin
    print(n,s,e,w)
    path = []
    Q2 = nx.MultiDiGraph()
    try:
        Q2 = ox.graph_from_bbox((w,s,e,n),simplify=False,retain_all=True,custom_filter=cf_add_routes)
        path = nx.shortest_path(Q2,node,node2,weight='length')
    except Exception as ex:
        print(ex)
        path = []
    result = False    
    box = (n,s,e,w)
    if len(path) > 0:
        T = Q2.subgraph(path).copy()
        Q = nx.compose(Q,T)
        print('added road across ',node,node2)
        result = True
    return result, Q, box    
            
# Because our program development needs a great deal of iteration,
# store intermediate results in pickle files
def find_trailheads(Q,QR,road_distance_limit = 30.0):
    trailheads_filename = 'trailheads.pkl'
    trailhead_nodes = set()
    if os.path.exists(trailheads_filename):
        trailhead_nodes = pickle.load(open(trailheads_filename, 'rb'))
        print('finished loading trailhead nodes')
        elapsed_time(start_time)
    else:
        for node in Q.nodes:
            d = Q.degree(node)
            if d == 2: #Q is multidigraph, so this is an end-point of trail
                if distance_road(Q,node,QR) < road_distance_limit :
                    trailhead_nodes.add(node)
                    print('added node to trailhead list: ', node)
                    continue
        pickle.dump(trailhead_nodes,open(trailheads_filename, 'wb'))
        print('finished calculating trailhead nodes')
        elapsed_time(start_time)
    return trailhead_nodes

def find_trailhead_pairs(Q, trailhead_nodes, distance_limit = 2000.0):
    trailhead_pairs = set()
    trailhead_pairs_filename = 'trailhead_pairs.pkl'
    if os.path.exists(trailhead_pairs_filename):
        trailhead_pairs = pickle.load(open(trailhead_pairs_filename, 'rb'))
        print('finished loading trailhead pairs')
    else:            
        for node in trailhead_nodes:
            for node2 in trailhead_nodes:
                if node==node2:
                    continue
                if (min(node,node2),max(node,node2)) in trailhead_pairs:
                    continue
                d= distance(Q, node,node2)
                #print('distance between ',node, node2, ' = ', d)
                if d < distance_limit:
                    trailhead_pairs.add((min(node, node2),max(node, node2)))
                    print('trailhead pairs: ', node, node2)
        pickle.dump(trailhead_pairs,open(trailhead_pairs_filename, 'wb'))
        print('finished calculating trailhead pairs')
    elapsed_time(start_time)
    print('length of trailhead_pairs: ',len(trailhead_pairs))
    return trailhead_pairs            


def mind_the_gap(Q,QR,distance_limit = 2000.0):
    print('start mind_the_gap function')
    
    trailhead_nodes = set()
    trailhead_nodes = find_trailheads(Q,QR)
    print('number of trailhead nodes: ', len(trailhead_nodes))
    
    trailhead_pairs = find_trailhead_pairs(Q,trailhead_nodes,distance_limit)

    joined_trailhead_boxes = []
    not_joined_trailhead_boxes = []
    for node, node2 in trailhead_pairs:
        result, QT, box = connect_trailheads_with_road(Q,node,node2)
        if result:
            joined_trailhead_boxes.append(box)
            Q = QT
        else:
            not_joined_trailhead_boxes.append(box)
    
    return Q, joined_trailhead_boxes, not_joined_trailhead_boxes
        

G, boxes, not_boxes = mind_the_gap(G,GR)  # has to be done on multidigraph, not multigraph
print('length of added boxes: ', len(boxes))
print('length of not-added boxes: ', len(not_boxes))

if len(boxes) > 0:
    pickle.dump(G,open('g_connected_gaps.pkl', 'wb'))
    H = ox.convert.to_undirected(G)
    draw(H,'Gila with connected gaps',boxes = boxes, not_boxes = not_boxes)
    pickle.dump(H,open('h_connected_gaps.pkl', 'wb'))
    
elapsed_time(start_time)
print('finished script ' + os.path.basename(__file__))






